// Chris Gerlinsky, 2024-01-25

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>


#define MAX_PATCH_LEN   1024

#define BUFFER_LEN      16384


// convert hex digits in hexstr to bytes in binary[]
int string_to_hex(const char *hexstr, unsigned char *binary, int maxlen)
{
    int DataByte;
    char ReadByte;
    int i;
    int nibble = 0;
    int binary_len = 0;
    int reading_string = 0;
    int escaped = 0;


    if( !isxdigit(hexstr[0]) )
        reading_string = 1;     // if the first byte of the string is not a hex number then read as string

    for( i=0; i<strlen(hexstr); i++ )
    {
        ReadByte = hexstr[i];
        
        if( escaped )
        {
            switch( ReadByte )
            {
                case '\n':
                    binary[binary_len++] = '\n';
                    break;
                    
                case '\r':
                    binary[binary_len++] = '\r';
                    break;
                    
                default:
                    binary[binary_len++] = ReadByte;
                    break;
            }
            escaped = 0;
        }
        
        if( ReadByte == '\"' && !reading_string )
        {
            reading_string = 1;
            continue;
        }
        else if( ReadByte == '\"' && reading_string )    // reached the end of string
        {
            if( i )
                reading_string = 0;
        }
        else if( ReadByte == '\\' && reading_string )
            escaped = 1;
        else if( reading_string )
            binary[binary_len++] = ReadByte;

        if( reading_string )
            continue;            


        if( nibble == 0 )
            DataByte = 0;
        else
            DataByte *= 0x10;

        if( ReadByte >= '0' && ReadByte <= '9')
            DataByte += ReadByte-'0';
        else if( ReadByte >= 'A' && ReadByte <= 'F')
            DataByte += ReadByte-'A'+ 0x0A;
        else if( ReadByte >= 'a' && ReadByte <= 'f')
            DataByte += ReadByte - 'a' + 0x0A;
        else if( nibble == 0 )
            continue;
        else if( nibble == 1 && (ReadByte == 'x' || ReadByte == 'X' ) )
        {   // we read in "0x" prefix, throw it away
            nibble = 0;
            continue;
        }
        else
            DataByte /= 0x10;       // second nibble wasn't read, so adjust databyte to put nibble we did read in low bits
        nibble++;
        
        if( nibble == 2 )
        {
            binary[binary_len++] = DataByte;
            nibble = 0;
            if( binary_len >= maxlen )
                break;
        }
    }
    if( nibble == 1 )
        binary[binary_len++] = DataByte;


    return binary_len;
}


int main(int argc,char **argv)
{
    FILE *InFile = NULL;
    char InFilename[FILENAME_MAX];
    unsigned char pattern[MAX_PATCH_LEN];
    int pattern_len = 0;
    unsigned char patch[MAX_PATCH_LEN];
    int patch_len = 0;
    int i;
    int verbose = 1;
    int write_patched_file = 1;
    long file_len = 0;
    int searchidx;
    long fileidx;
    long match_address;
    unsigned char *buffer = NULL;
    int buffer_len = 0;
    int bytes_in_buffer = 0;
    int bufidx;
    int match_count = 0;
    int patch_count = 0;
    long cur_file_pos = 0;
    

    if( argc < 4 )
    {
        printf("usage: %s <pattern> <patch> <filename> [options]\n",argv[0]);
        printf("\n");
        printf("options:\n");
        printf("\t-v\tverbose (increase verbosity)\n");
        printf("\t-q\tquiet (reduce verbosity)\n");
        printf("\t-n\tdo not write patch to file\n");
        printf("\n");
        printf("If first character of <pattern> is hex digit then it's interpeted as hex bytes,\n");
        printf("Otherwise it is interpreted as an ASCII string.\n");
        printf("Double quotes \" \" can be used to specify an ASCII string.\n");
        
        
        return 0;
    }

    for( i=4; i<argc; i++ )
    {
        if( argv[i][0] == '-' )
        {
            switch( tolower(argv[i][1]) )
            {
                case 'v':
                    verbose++;
                    break;
                
                case 'q':
                    verbose--;
                    break;
                
                case 'n':
                    write_patched_file--;
                    break;
            }
        }
    }

    if( verbose > 1 )
    {
        printf( "verbose: %d\n", verbose );
        printf( "write_patched_file: %d\n", write_patched_file );
    }

    buffer = (unsigned char *)malloc(BUFFER_LEN);
    if( buffer )
    {
        buffer_len = BUFFER_LEN;
        if( verbose > 1 )
            printf( "allocated %d bytes for buffer.\n", buffer_len );
    }
    else
    {
        printf( "error allocating %d bytes for buffer - aborting.\n", BUFFER_LEN );
        return 0;
    }

    strcpy(InFilename,argv[3]);
    InFile=fopen(InFilename,"r+b");
    if(!InFile)
    {
        printf("error opening file '%s'\n",InFilename);
        return 0;
    }

    pattern_len = string_to_hex(argv[1], pattern, MAX_PATCH_LEN);
    patch_len = string_to_hex(argv[2], patch, MAX_PATCH_LEN);

    if( verbose >= 1 )
    {
        printf( "pattern:" );
        for( i=0; i<pattern_len; i++ )
        {
            if( i && (!i%32) )
                printf( "\n" );
            printf( " %02X", pattern[i] );
        }
        printf( "\n" );
    
        printf( "patch:  " );
        for( i=0; i<patch_len; i++ )
        {
            if( i && (!i%32) )
                printf( "\n" );
            printf( " %02X", patch[i] );
        }
        printf( "\n" );
        
        if( pattern_len != patch_len )
        {
            printf( "WARNING: patch length (%d) doesn't match pattern length (%d).\n", patch_len, pattern_len );
            printf( "Patch will be applied over the beginning of the matched pattern.\n" );
        }
    }

    fseek( InFile, 0, SEEK_END );    
    file_len = ftell( InFile );    
    fseek( InFile, 0, SEEK_SET );

    if( verbose > 1 )
        printf( "file length: %ld\n", file_len );


    fileidx = 0;
    searchidx = 0;

    while ( !feof(InFile) && !ferror(InFile) )
    {
        if( searchidx )
        {   // we are partway through matching a pattern but reached end of buffer
            // move the pattern that has been matched so far to the beginning of buffer
            memmove( buffer, buffer+bufidx-searchidx, searchidx );

            bufidx += bytes_in_buffer;
            bufidx -= searchidx;    // adjust bufidx to compensate for leaving bytes in the buffer
        }

        bytes_in_buffer = fread( buffer+searchidx, 1, buffer_len-searchidx, InFile );
        if( verbose > 1 )
            printf( "read %d bytes to buffer[]\n", bytes_in_buffer );

        for( bufidx=searchidx; bufidx<bytes_in_buffer; bufidx++ )
        {            
            if( buffer[bufidx] == pattern[searchidx] )
            {
                searchidx++;
            
                if( searchidx == pattern_len )
                {   // entire pattern was matched!
                    match_address = fileidx + bufidx - pattern_len + 1;
                    
                    match_count++;
                
                    if( verbose > 0 )
                        printf( "Found @ %lX ", match_address );
                    
                    searchidx = 0;

                    if( !write_patched_file )
                    {
                        if( verbose > 0 )
                            printf( "\n" );
                        continue;
                    }

                    cur_file_pos = ftell(InFile);
                    fseek( InFile, match_address, SEEK_SET );
                    
                    for( i=0; i<patch_len; i++ )
                    {
                        if( fputc( patch[i], InFile ) != patch[i] )
                        {
                            printf( "ERROR: can't write file.\n" );
                            break;
                        }
                    }
                    if( i >= patch_len )
                    {
                        patch_count++;
                        
                        if( verbose > 0 )
                            printf("... Patched %d bytes.\n", patch_len);                           
                    }

                    fseek( InFile, cur_file_pos, SEEK_SET );
                }
            }
            else if( searchidx )
            {
                searchidx--;
                bufidx -= searchidx;    // adjust bufidx to search for pattern from the byte after previous false start

                searchidx = 0;
            }
        }
        
        fileidx += bufidx;
    }


    if( InFile )
        fclose(InFile);

    if( verbose > 0 )
    {
        if( match_count != patch_count || verbose > 1 )
            printf("Done: %d matches, %d patches applied.\n", match_count, patch_count);
        else
            printf("Done: %d patches applied.\n", patch_count);
    }


    return 0;
}
